class CustomerService(object):
    """
    :class:`fortnox.CustomerService` is used by :class:`fortnox.Client` to make
    actions related to Customer resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for Customer to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = ['Name']
    SERVICE = "Customer"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def list(self, **params):
        """
        Retrieve all customers

        Returns all Customers available to the Company, according to the parameters provided

        :calls: ``get /customers``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of Customers.
        :rtype: list
        """

        _, _, customers = self.http_client.get("/customers", params=params)
        return customers

    def retrieve(self, id):
        """
        Retrieve a single customer

        Returns a single Customer according to the unique Customer ID provided
        If the specified Customer does not exist, this query returns an error

        :calls: ``get /customers/{id}``
        :param int id: Unique identifier of a Customer.
        :return: Dictionary that support attriubte-style access and represent User resource.
        :rtype: dict
        """
        _, _, customer = self.http_client.get("/customers/{id}".format(id=id))
        return customer

    def create(self, *args, **kwargs):
        """
        Create a customer

        Creates a new customer
        **Notice** the customer's name **must** be unique within the scope of the resource_type

        :calls: ``post /customers``
        :param tuple *args: (optional) Single object representing Customer resource.
        :param dict **kwargs: (optional) Customer attributes.
        :return: Dictionary that support attriubte-style access and represents newely created Customer resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Customer are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items() if k in self.OPTS_KEYS_TO_PERSIST)
        attributes.update({'service': self.SERVICE})
        _, _, customer = self.http_client.post("/customers", body=attributes)
        return customer

    def update(self, id, *args, **kwargs):
        """
        Update a customer

        Updates a customer's information
        If the specified customer does not exist, this query will return an error
        **Notice** if you want to update a customer, you **must** make sure the customer's name is unique within the scope of the specified resource

        :calls: ``put /customers/{id}``
        :param int id: Unique identifier of a Customer.
        :param tuple *args: (optional) Single object representing Customer resource which attributes should be updated.
        :param dict **kwargs: (optional) Customer attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated Customer resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Customer are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, customer = self.http_client.put("/customers/{id}".format(id=id), body=attributes)
        return customer

    def destroy(self, id):
        """
        Delete a customer

        Deletes an existing customer
        If the specified customer is assigned to any resource, we will remove this customer from all such resources
        If the specified customer does not exist, this query will return an error
        This operation cannot be undone

        :calls: ``delete /customers/{id}``
        :param int id: Unique identifier of a Customer.
        :return: True if the operation succeeded.
        :rtype: bool
        """

        status_code, _, _ = self.http_client.delete("/customers/{id}".format(id=id))
        return status_code == 204



class PredefinedAccountService(object):
    """
    :class:`fortnox.PredefinedAccountService` is used by :class:`fortnox.Client` to make
    actions related to Predefined Account resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for PredefinedAccount to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = []
    SERVICE = "PreDefinedAccount"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def list(self, **params):
        """
        Retrieve all Predefined Accounts

        Returns all Predefined Accounts available to the Company, according to the parameters provided

        :calls: ``get /predefinedaccounts``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of Predefined Accounts.
        :rtype: list
        """

        _, _, predefined_accounts = self.http_client.get("/predefinedaccounts", params=params)
        return predefined_accounts

    def retrieve(self, name):
        """
        Retrieve a single Predefined Account

        Returns a single Predefined Account according to the unique Predefined Account ID provided
        If the specified Predefined Account does not exist, this query returns an error

        :calls: ``get /predefinedaccounts/{name}``
        :param int id: Unique identifier of a Predefined Account.
        :return: Dictionary that support attriubte-style access and represent Predefined Account resource.
        :rtype: dict
        """
        _, _, predefined_account = self.http_client.get("/predefinedaccounts/{name}".format(name=name))
        return predefined_account

    def update(self, name, *args, **kwargs):
        """
        Update a Predefined Account

        Updates a Predefined Account's information
        If the specified Predefined Account does not exist, this query will return an error
        **Notice** if you want to update a Predefined Account, you **must** make sure the Predefined Account's name is unique within the scope of the specified resource

        :calls: ``put /predefinedaccounts/{name}``
        :param int id: Unique identifier of a Predefined Account.
        :param tuple *args: (optional) Single object representing Predefined Account resource which attributes should be updated.
        :param dict **kwargs: (optional) Predefined Account attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated Predefined Account resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for PredefinedAccount are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, predefined_account = self.http_client.put("/predefinedaccounts/{name}".format(name=name), body=attributes)
        return predefined_account



class FinancialYearService(object):
    """
    :class:`fortnox.FinancialYearService` is used by :class:`fortnox.Client` to make
    actions related to FinancialYear resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for FinancialYear to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = ['FromDate', 'ToDate', 'AccountChartType']
    SERVICE = "FinancialYear"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def list(self, **params):
        """
        Retrieve all FinancialYears

        Returns all FinancialYears available to the Company, according to the parameters provided

        :calls: ``get /financialyears``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of FinancialYears.
        :rtype: list
        """

        _, _, financial_years = self.http_client.get("/financialyears", params=params)
        return financial_years

    def retrieve(self, id):
        """
        Retrieve a single FinancialYear

        Returns a single FinancialYear according to the unique FinancialYear ID provided
        If the specified FinancialYear does not exist, this query returns an error

        :calls: ``get /financialyears/{id}``
        :param int id: Unique identifier of a FinancialYear.
        :return: Dictionary that support attriubte-style access and represent FinancialYear resource.
        :rtype: dict
        """
        _, _, financial_year = self.http_client.get("/financialyears/{id}".format(id=id))
        return financial_year

    def create(self, *args, **kwargs):
        """
        Create a FinancialYear

        Creates a new FinancialYear
        **Notice** the FinancialYear's name **must** be unique within the scope of the resource_type

        :calls: ``post /financialyears``
        :param tuple *args: (optional) Single object representing FinancialYear resource.
        :param dict **kwargs: (optional) FinancialYear attributes.
        :return: Dictionary that support attriubte-style access and represents newely created FinancialYear resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for FinancialYear are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items() if k in self.OPTS_KEYS_TO_PERSIST)
        attributes.update({'service': self.SERVICE})
        _, _, financial_year = self.http_client.post("/financialyears", body=attributes)
        return financial_year



from fortnox.errors import (
    ConfigurationError, RateLimitError, BaseError,
    RequestError, ResourceError, ServerError
)

from fortnox.configuration import Configuration
from fortnox.http_client import HttpClient

from fortnox.services import (
    AbsenceTransactionsService, CurrencyService, CompanyInformationService, ProjectService,
    CustomerService, CompanySettingsService, AccountChartsService, AccountsService,
    FinancialYearService, InvoiceService, AssetTypeService, PredefinedAccountService,
    EmployeeService, ArticleService, ExpenseService, PredefinedVoucherSeriesService,
    AssetService,
)
from fortnox.client import Client



import json
from decimal import Decimal

import requests
from munch import munchify

from fortnox.errors import ResourceError, RateLimitError, RequestError, ServerError


class DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            return float(o)
        super(DecimalEncoder, self).default(o)


class HttpClient(object):
    """
    Wrapper over :module:`requests` that understands Base CRM envelope, encoding and decoding schema.
    """

    """
    Supported REST API version prefix.
    """
    API_VERSION = '/3'

    def __init__(self, config):
        """
        :param :class:`fortnox.Configuration` config: Base CRM client configuration.
        """

        self.config = config

        # if self.config.verbose:
        #     self.enable_logging()

    def get(self, url, params=None, **kwargs):
        """
        Send a GET request.

        :param str url: Sub URL for the request. You MUST not specify neither base url nor api version prefix.
        :param dict params: (optional) Dictionary of query parameters.
        :param dict **kwargs: (optional) Other parameters which are directly passed to :func:`requests.request`.
        :return: Tuple of three elements: (http status code, headers, response - either parsed json or plain text)
        :rtype: tuple
        """

        return self.request('get', url, params=params, **kwargs)

    def post(self, url, body=None, **kwargs):
        """
        Send a POST request.

        :param str url: Sub URL for the request. You MUST not specify neither base url nor api version prefix.
        :param dict body: (optional) Dictionary of body attributes that will be wrapped with envelope and json encoded.
        :param dict **kwargs: (optional) Other parameters which are directly passed to :func:`requests.request`.
        :return: Tuple of three elements: (http status code, headers, response - either parsed json or plain text)
        :rtype: tuple
        """

        return self.request('post', url, body=body, **kwargs)

    def put(self, url, body=None, **kwargs):
        """
        Send a PUT request.

        :param str url: Sub URL for the request. You MUST not specify neither base url nor api version prefix.
        :param dict body: (optional) Dictionary of body attributes that will be wrapped with envelope and json encoded.
        :param dict **kwargs: (optional) Other parameters which are directly passed to :func:`requests.request`.
        :return: Tuple of three elements: (http status code, headers, response - either parsed json or plain text)
        :rtype: tuple
        """

        return self.request('put', url, body=body, **kwargs)

    def delete(self, url, params=None, **kwargs):
        """
        Send a DELETE request.

        :param str url: Sub URL for the request. You MUST not specify neither base url nor api version prefix.
        :param dict params: (optional) Dictionary of query parameters.
        :param dict **kwargs: (optional) Other parameters which are directly passed to :func:`requests.request`.
        :return: Tuple of three elements: (http status code, headers, response - either parsed json or plain text)
        :rtype: tuple
        """

        return self.request('delete', url, params=params, **kwargs)

    def request(self, method, url, params=None, body=None, **kwargs):
        """
        Send an HTTP request.

        The :param:`params` will be properly encoded, as well as :param:`body` which
        will be wrapped with envelope the API expects and json encoded.

        When you get a response the method will try to json decode the response,
        if the media type represents json, unwrap the envelope and munchify what has left,
        for JavaScript like access.

        :param str url: Sub URL for the request. You MUST not specify neither base url nor api version prefix.
        :param dict params: (optional) Dictionary of query parameters.
        :param dict body: (optional) Dictionary of body attributes that will be wrapped with envelope and json encoded.
        :param dict **kwargs: (optional) Other parameters which are directly passed to :func:`requests.request`.
        :raises RequestError: if authentication failed, invalid query parameter etc.
        :raises RateLimitError: if rate limit exceeded.
        :raises ResourceError: if requests payload included invalid attributes or were missing.
        :raises ServerError: if Base CRM backend servers encounterered an unexpected condition.
        :return: Tuple of three elements: (http status code, headers, response - either parsed json or plain text)
        :rtype: tuple

        :Keyword Arguments:
            * :param dict headers: (optional) Dictionary of headers. Default: ``{}``.
            * :param bool raw: (optional) Whether to wrap and uwrap the envelope. Default: ``False``.
        """

        url = "{base_url}{version}{resource}".format(base_url=self.config.base_url,
                                                     version=self.API_VERSION,
                                                     resource=url)
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Access-Token': "{access_token}".format(access_token=self.config.access_token),
            'Client-Secret': "{client_secret}".format(client_secret=self.config.client_secret)
        }

        user_headers = {}
        if 'headers' in kwargs and isinstance(kwargs['headers'], dict):
            user_headers = kwargs['headers']

        headers.update(user_headers)

        raw = bool(kwargs['raw']) if 'raw' in kwargs else False

        if body is not None:
            # payload = body if raw else self.wrap_envelope(body)
            body = json.dumps(self.wrap_envelope(body), cls=DecimalEncoder)
        resp = requests.request(method, url,
                                params=params,
                                data=body,
                                headers=headers,
                                timeout=float(self.config.timeout),
                                )

        if not (200 <= resp.status_code < 300):
            self.handle_error_response(resp)

        if 'Content-Type' in resp.headers and 'json' in resp.headers['Content-Type']:
            resp_body = munchify(resp.json()) if raw else self.unwrap_envelope(resp.json())
        else:
            resp_body = resp.content

        return (resp.status_code, resp.headers, resp_body)

    def handle_error_response(self, resp):
        try:
            errors = resp.json()
        except ValueError:
            errors = {'errors': [], 'meta': {'logref': ''}}

        resp_code = resp.status_code
        if resp_code == 422:
            raise ResourceError(resp_code, errors)
        elif resp_code == 429:
            raise RateLimitError()
        elif 400 <= resp_code < 500:
            raise RequestError(resp_code, errors)
        elif 500 <= resp_code < 600:
            raise ServerError(resp_code, errors)
        else:
            raise Exception('Unknown HTTP error response')

    @staticmethod
    def wrap_envelope(body):
        key = 'data'
        if 'service' in body.keys():
            key = body.get('service')
            body.pop('service')
        return {key: body}

    @staticmethod
    def unwrap_envelope(body):
        if body is '' or body is None:
            # Sometimes response body returns nothing from Fortnox API
            return True
        keys = [key for key in body.keys()]
        return [munchify(item) for item in body.get(keys[1])] if len(keys) > 1 else munchify(body.get(keys[0]))

    def enable_logging(self):
        import logging
        try:
            import http.client as http_client
        except ImportError:
            # Python 2
            import httplib as http_client
        http_client.HTTPConnection.debuglevel = 1

        logging.basicConfig()
        logging.getLogger().setLevel(logging.DEBUG)
        requests_log = logging.getLogger("requests.packages.urllib3")
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True



class AssetTypeService(object):
    """
    :class:`fortnox.AssetTypeService` is used by :class:`fortnox.Client` to make
    actions related to Asset Type resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for AssetType to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = ['Number', 'Description', 'AccountAssetId',
                            'AccountDepreciationId', 'AccountValueLossId',
                            'Type']
    SERVICE = "AssetType"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def list(self, **params):
        """
        Retrieve all AssetTypes

        Returns all AssetTypes available to the Company, according to the parameters provided

        :calls: ``get /assets/types``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of AssetTypes.
        :rtype: list
        """

        _, _, asset_types = self.http_client.get("/assets/types", params=params)
        return asset_types[1:]

    def retrieve(self, id):
        """
        Retrieve a single AssetType

        Returns a single AssetType according to the unique AssetType ID provided
        If the specified AssetType does not exist, this query returns an error

        :calls: ``get /assets/types/{id}``
        :param int id: Unique identifier of an AssetType.
        :return: Dictionary that support attriubte-style access and represent AssetType resource.
        :rtype: dict
        """
        _, _, asset_type = self.http_client.get("/assets/types/{id}".format(id=id))
        return asset_type

    def create(self, *args, **kwargs):
        """
        Create an AssetType

        Creates a new AssetType
        **Notice** the AssetType's name **must** be unique within the scope of the resource_type

        :calls: ``post /assets/types``
        :param tuple *args: (optional) Single object representing AssetType resource.
        :param dict **kwargs: (optional) AssetType attributes.
        :return: Dictionary that support attriubte-style access and represents newely created AssetType resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for AssetType are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items() if k in self.OPTS_KEYS_TO_PERSIST)
        attributes.update({'service': self.SERVICE})
        _, _, asset_type = self.http_client.post("/assets/types", body=attributes)
        return asset_type

    def update(self, id, *args, **kwargs):
        """
        Update an AssetType

        Updates an AssetType's information
        If the specified AssetType does not exist, this query will return an error
        **Notice** if you want to update an AssetType, you **must** make sure the AssetType's name is unique within the scope of the specified resource

        :calls: ``put /assets/types/{id}``
        :param int id: Unique identifier of an AssetType.
        :param tuple *args: (optional) Single object representing AssetType resource which attributes should be updated.
        :param dict **kwargs: (optional) AssetType attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated AssetType resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for AssetType are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, asset_type = self.http_client.put("/assets/types/{id}".format(id=id), body=attributes)
        return asset_type

    def destroy(self, id):
        """
        Delete an AssetType

        Deletes an existing AssetType
        If the specified AssetType is assigned to any resource, we will remove this AssetType from all such resources
        If the specified AssetType does not exist, this query will return an error
        This operation cannot be undone

        :calls: ``delete /assets/types/{id}``
        :param int id: Unique identifier of an AssetType.
        :return: True if the operation succeeded.
        :rtype: bool
        """

        status_code, _, _ = self.http_client.delete("/assets/types/{id}".format(id=id))
        return status_code == 200



class ArticleService(object):
    """
    :class:`fortnox.ArticleService` is used by :class:`fortnox.Client` to make
    actions related to Articles resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for Articles to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = ['ArticleNumber', 'Description']
    SERVICE = "Article"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def list(self, **params):
        """
        Retrieve all Articles

        Returns all Articles available to the Company, according to the parameters provided

        :calls: ``get /articles``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of Articles.
        :rtype: list
        """

        _, _, articles = self.http_client.get("/articles", params=params)
        return articles

    def retrieve(self, number):
        """
        Retrieve a single Article

        Returns a single Article according to the unique Article ID provided
        If the specified Article does not exist, this query returns an error

        :calls: ``get /articles/{number}``
        :param int id: Unique identifier of an Article.
        :return: Dictionary that support attriubte-style access and represent Article resource.
        :rtype: dict
        """
        _, _, article = self.http_client.get("/articles/{number}".format(number=number))
        return article

    def create(self, *args, **kwargs):
        """
        Create an Article

        Creates a new Article
        **Notice** the Article's name **must** be unique within the scope of the resource_type

        :calls: ``post /articles``
        :param tuple *args: (optional) Single object representing Article resource.
        :param dict **kwargs: (optional) Article attributes.
        :return: Dictionary that support attriubte-style access and represents newely created Articles resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Article are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items() if k in self.OPTS_KEYS_TO_PERSIST)
        attributes.update({'service': self.SERVICE})
        _, _, article = self.http_client.post("/articles", body=attributes)
        return article

    def update(self, number, *args, **kwargs):
        """
        Update an Article

        Updates an Article's information
        If the specified Article does not exist, this query will return an error
        **Notice** if you want to update an Article, you **must** make sure the Article's name is unique within the scope of the specified resource

        :calls: ``put /articles/{number}``
        :param int id: Unique identifier of an Article.
        :param tuple *args: (optional) Single object representing Article resource which attributes should be updated.
        :param dict **kwargs: (optional) Article attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated Article resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Article are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, article = self.http_client.put("/articles/{number}".format(number=number), body=attributes)
        return article

    def destroy(self, number):
        """
        Delete an Article

        Deletes an existing Article
        If the specified Article is assigned to any resource, we will remove this Article from all such resources
        If the specified Article does not exist, this query will return an error
        This operation cannot be undone

        :calls: ``delete /articles/{number}``
        :param int id: Unique identifier of an Article.
        :return: True if the operation succeeded.
        :rtype: bool
        """

        status_code, _, _ = self.http_client.delete("/articles/{number}".format(number=number))
        return status_code == 204



class Service(object):
    """
    :class:`fortnox.<specific-service>` is used by :class:`fortnox.Client` to make
    actions related to <specific-service> resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for <specific-service> to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = ['Name']
    SERVICE = "<specific-service>"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def list(self, **params):
        """
        Retrieve all <specific-service>

        Returns all <specific-service> available to the Company, according to the parameters provided

        :calls: ``get /<specific-service>``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of Customers.
        :rtype: list
        """

        _, _, customers = self.http_client.get("/<specific-service>", params=params)
        return customers

    def retrieve(self, id):
        """
        Retrieve a single <specific-service>

        Returns a single <specific-service> according to the unique <specific-service> ID provided
        If the specified <specific-service> does not exist, this query returns an error

        :calls: ``get /<specific-service>/{id}``
        :param int id: Unique identifier of a <specific-service>.
        :return: Dictionary that support attriubte-style access and represent <specific-service> resource.
        :rtype: dict
        """
        _, _, customer = self.http_client.get("/<specific-service>/{id}".format(id=id))
        return customer

    def create(self, *args, **kwargs):
        """
        Create a <specific-service>

        Creates a new customer
        **Notice** the customer's name **must** be unique within the scope of the resource_type

        :calls: ``post /customers``
        :param tuple *args: (optional) Single object representing <specific-service> resource.
        :param dict **kwargs: (optional) Customer attributes.
        :return: Dictionary that support attriubte-style access and represents newely created Customer resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for <specific-service> are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items() if k in self.OPTS_KEYS_TO_PERSIST)
        attributes.update({'service': self.SERVICE})
        _, _, customer = self.http_client.post("/<specific-service>", body=attributes)
        return customer

    def update(self, id, *args, **kwargs):
        """
        Update a <specific-service>

        Updates a <specific-service>'s information
        If the specified <specific-service> does not exist, this query will return an error
        **Notice** if you want to update a <specific-service>, you **must** make sure the <specific-service>'s name is unique within the scope of the specified resource

        :calls: ``put /<specific-service>/{id}``
        :param int id: Unique identifier of a <specific-service>.
        :param tuple *args: (optional) Single object representing <specific-service> resource which attributes should be updated.
        :param dict **kwargs: (optional) <specific-service> attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated <specific-service> resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for <specific-service> are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, customer = self.http_client.put("/customers/{id}".format(id=id), body=attributes)
        return customer

    def destroy(self, id):
        """
        Delete a <specific-service>

        Deletes an existing <specific-service>
        If the specified <specific-service> is assigned to any resource, we will remove this <specific-service> from all such resources
        If the specified <specific-service> does not exist, this query will return an error
        This operation cannot be undone

        :calls: ``delete /<specific-service>/{id}``
        :param int id: Unique identifier of a <specific-service>.
        :return: True if the operation succeeded.
        :rtype: bool
        """

        status_code, _, _ = self.http_client.delete("/<specific-service>/{id}".format(id=id))
        return status_code == 204



class CompanySettingsService(object):
    """
    :class:`fortnox.CompanySettingsService` is used by :class:`fortnox.Client` to make
    actions related to CompanySettings resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for CompanySettings to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = []
    SERVICE = "CompanySettings"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def info(self, **params):
        """
        Retrieve CompanySettings

        Returns all information available to the CompanySettings, according to the parameters provided

        :calls: ``get /settings/company``
        :param dict params: (optional) Search options.
        :return: dictionaries that support attriubte-style access, which represent collection of CompanySettings.
        :rtype: munch object
        """

        _, _, company_settings = self.http_client.get("/settings/company", params=params)
        return company_settings



from .customer_services import CustomerService
from .company_settings_services import CompanySettingsService
from .account_chart_services import AccountChartsService
from .account_services import AccountsService
from .absence_transaction_services import AbsenceTransactionsService
from .currency_services import CurrencyService
from .company_information_services import CompanyInformationService
from .employee_services import EmployeeService
from .project_services import ProjectService
from .article_services import ArticleService
from .expense_services import ExpenseService
from .financial_year_services import FinancialYearService
from .invoice_services import InvoiceService
from .asset_type_services import AssetTypeService
from .predefined_account_services import PredefinedAccountService
from .predefined_voucher_series_services import PredefinedVoucherSeriesService
from .asset_services import AssetService



from fortnox.configuration import Configuration
from fortnox.http_client import HttpClient

import fortnox.services


class Client(object):
    """
    The :class:`Client <Client>` is the entry point to all services and actions.

    :attribute :class:`Configuration <fortnox.Configuration>` config: Current Fortnox client configuration.
    :attribute :class:`HttpClient <fortnox.HttpClient>` http_client: Http client.

    :copyright: (c) 2020 by Mahmudul Hasan (ikhtiarcse10ruet@gmail.com).
    :license: MIT, see LICENSE for more details.
    """

    def __init__(self, **options):
        """
        Usage::

          >>> import os
          >>> import fortnox
          >>> client = fortnox.Client(access_token=os.environ.get('ACCESS_TOKEN'), client_secret=os.environ.get('CLIENT_SECRET'))
          <fortnox.Client>


        :param str access_token: Personal access token.
        :param str client_secret: Personal access token.
        :param str base_url: (optional) Base url for the api. Default: ``https://api.getbase.com``.
        :param bool verbose: (optional) Verbose/debug mode. Default: ``False``.
        :param int timeout: (optional) Connection and response timeout. Default: **30** seconds.

        :raises ConfigurationError: if no ``access_token`` provided.
        :raises ConfigurationError: if provided ``access_token`` is invalid - contains disallowed characters.
        :raises ConfigurationError: if provided ``access_token`` is invalid - has invalid length.
        :raises ConfigurationError: if provided ``base_url`` is invalid.
        """

        self.config = Configuration(**options)
        self.config.validate()

        self.http_client = HttpClient(self.config)

        self.__customers = fortnox.services.CustomerService(self.http_client)
        self.__company_settings = fortnox.services.CompanySettingsService(self.http_client)
        self.__account_charts = fortnox.services.AccountChartsService(self.http_client)
        self.__accounts = fortnox.services.AccountsService(self.http_client)
        self.__absence_transactions = fortnox.services.AbsenceTransactionsService(self.http_client)
        self.__currencies = fortnox.services.CurrencyService(self.http_client)
        self.__company_information = fortnox.services.CompanyInformationService(self.http_client)
        self.__employees = fortnox.services.EmployeeService(self.http_client)
        self.__projects = fortnox.services.ProjectService(self.http_client)
        self.__articles = fortnox.services.ArticleService(self.http_client)
        self.__expenses = fortnox.services.ExpenseService(self.http_client)
        self.__financial_years = fortnox.services.FinancialYearService(self.http_client)
        self.__invoices = fortnox.services.InvoiceService(self.http_client)
        self.__asset_types = fortnox.services.AssetTypeService(self.http_client)
        self.__predefined_accounts = fortnox.services.PredefinedAccountService(self.http_client)
        self.__predefined_voucher_series = fortnox.services.PredefinedVoucherSeriesService(self.http_client)
        self.__assets = fortnox.services.AssetService(self.http_client)

    @property
    def customers(self):
        """
        :return: :class:`CustomerService <fortnox.CustomerService>` object that gives you an access to all Customers related actions.
        :rtype: fortnox.CustomerService
        """
        return self.__customers

    @property
    def company_settings(self):
        """
        :return: :class:`CompanySettingsService <fortnox.CompanySettingsService>` object that gives you an access to CompanySettings related actions.
        :rtype: fortnox.CompanySettingsService
        """
        return self.__company_settings

    @property
    def account_charts(self):
        """
        :return: :class:`AccountChartsService <fortnox.AccountChartsService>` object that gives you an access to AccountChartsService related actions.
        :rtype: fortnox.AccountChartsService
        """
        return self.__account_charts

    @property
    def accounts(self):
        """
        :return: :class:`AccountsService <fortnox.AccountsService>` object that gives you an access to AccountsService related actions.
        :rtype: fortnox.AccountsService
        """
        return self.__accounts

    @property
    def absence_transactions(self):
        """
        :return: :class:`AbsenceTransactionsService <fortnox.AbsenceTransactionsService>` object that gives you an access to AbsenceTransactionsService related actions.
        :rtype: fortnox.AbsenceTransactionsService
        """
        return self.__absence_transactions

    @property
    def currencies(self):
        """
        :return: :class:`CurrencyService <fortnox.CurrencyService>` object that gives you an access to CurrencyService related actions.
        :rtype: fortnox.CurrencyService
        """
        return self.__currencies

    @property
    def company_information(self):
        """
        :return: :class:`CompanyInformationService <fortnox.CompanyInformationService>` object that gives you an access to CompanyInformationService related actions.
        :rtype: fortnox.CompanyInformationService
        """
        return self.__company_information

    @property
    def employees(self):
        """
        :return: :class:`EmployeeService <fortnox.EmployeeService>` object that gives you an access to Employees related actions.
        :rtype: fortnox.EmployeeService
        """
        return self.__employees

    @property
    def projects(self):
        """
        :return: :class:`ProjectService <fortnox.ProjectService>` object that gives you an access to Projects related actions.
        :rtype: fortnox.ProjectService
        """
        return self.__projects

    @property
    def articles(self):
        """
        :return: :class:`ArticleService <fortnox.ArticleService>` object that gives you an access to Articles related actions.
        :rtype: fortnox.ArticleService
        """
        return self.__articles

    @property
    def expenses(self):
        """
        :return: :class:`ExpenseService <fortnox.ExpenseService>` object that gives you an access to Expense related actions.
        :rtype: fortnox.ExpenseService
        """
        return self.__expenses

    @property
    def financial_years(self):
        """
        :return: :class:`FinancialYearService <fortnox.FinancialYearService>` object that gives you an access to Financial Year related actions.
        :rtype: fortnox.FinancialYearService
        """
        return self.__financial_years

    @property
    def invoices(self):
        """
        :return: :class:`InvoiceService <fortnox.InvoiceService>` object that gives you an access to Invoice related actions.
        :rtype: fortnox.InvoiceService
        """
        return self.__invoices

    @property
    def asset_types(self):
        """
        :return: :class:`AssetTypeService <fortnox.AssetTypeService>` object that gives you an access to Asset Type related actions.
        :rtype: fortnox.AssetTypeService
        """
        return self.__asset_types

    @property
    def predefined_accounts(self):
        """
        :return: :class:`PredefinedAccountService <fortnox.PredefinedAccountService>` object that gives you an access to Predefined Accounts related actions.
        :rtype: fortnox.PredefinedAccountService
        """
        return self.__predefined_accounts

    @property
    def predefined_voucher_series(self):
        """
        :return: :class:`PredefinedVoucherSeriesService <fortnox.PredefinedVoucherSeriesService>` object that gives you an access to Predefined Voucher Series related actions.
        :rtype: fortnox.PredefinedVoucherSeriesService
        """
        return self.__predefined_voucher_series

    @property
    def assets(self):
        """
        :return: :class:`AssetService <fortnox.AssetService>` object that gives you an access to Asset related actions.
        :rtype: fortnox.AssetService
        """
        return self.__assets



class ExpenseService(object):
    """
    :class:`fortnox.ExpenseService` is used by :class:`fortnox.Client` to make
    actions related to Expense resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for Expense to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = ['Code', 'Text', 'Account']
    SERVICE = "Expense"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def list(self, **params):
        """
        Retrieve all Expenses

        Returns all Expenses available to the Company, according to the parameters provided

        :calls: ``get /expenses``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of Expenses.
        :rtype: list
        """

        _, _, expenses = self.http_client.get("/expenses", params=params)
        return expenses

    def retrieve(self, expense_code):
        """
        Retrieve a single Expense

        Returns a single Expense according to the unique Expense ID provided
        If the specified Expense does not exist, this query returns an error

        :calls: ``get /expenses/{expense_code}``
        :param int id: Unique identifier of an Expense.
        :return: Dictionary that support attriubte-style access and represent Expense resource.
        :rtype: dict
        """
        _, _, expense = self.http_client.get("/expenses/{expense_code}".format(expense_code=expense_code))
        return expense

    def create(self, *args, **kwargs):
        """
        Create an Expense

        Creates a new Expense
        **Notice** the Expense's name **must** be unique within the scope of the resource_type

        :calls: ``post /expenses``
        :param tuple *args: (optional) Single object representing Expense resource.
        :param dict **kwargs: (optional) Expenses attributes.
        :return: Dictionary that support attriubte-style access and represents newely created Expense resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Expense are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items() if k in self.OPTS_KEYS_TO_PERSIST)
        attributes.update({'service': self.SERVICE})
        _, _, expense = self.http_client.post("/expenses", body=attributes)
        return expense



class InvoiceService(object):
    """
    :class:`fortnox.InvoiceService` is used by :class:`fortnox.Client` to make
    actions related to InvoiceService resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for Invoice to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = ['InvoiceRows', 'CustomerNumber']
    SERVICE = "Invoice"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def list(self, **params):
        """
        Retrieve all Invoices

        Returns all Invoices available to the Company, according to the parameters provided

        :calls: ``get /invoices``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of Invoices.
        :rtype: list
        """

        _, _, invoices = self.http_client.get("/invoices", params=params)
        return invoices

    def retrieve(self, id):
        """
        Retrieve a single Invoice

        Returns a single Invoice according to the unique Invoice ID provided
        If the specified Invoice does not exist, this query returns an error

        :calls: ``get /invoices/{id}``
        :param int id: Unique identifier of an Invoice.
        :return: Dictionary that support attriubte-style access and represent Invoice resource.
        :rtype: dict
        """
        _, _, invoice = self.http_client.get("/invoices/{id}".format(id=id))
        return invoice

    def create(self, *args, **kwargs):
        """
        Create an Invoice

        Creates a new Invoice
        **Notice** the invoice's name **must** be unique within the scope of the resource_type

        :calls: ``post /invoices``
        :param tuple *args: (optional) Single object representing Invoice resource.
        :param dict **kwargs: (optional) Invoice attributes.
        :return: Dictionary that support attriubte-style access and represents newely created Invoice resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Invoice are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items() if k in self.OPTS_KEYS_TO_PERSIST)
        attributes.update({'service': self.SERVICE})
        _, _, invoice = self.http_client.post("/invoices", body=attributes)
        return invoice

    def update(self, id, *args, **kwargs):
        """
        Update an Invoice

        Updates an invoice's information
        If the specified Invoice does not exist, this query will return an error
        **Notice** if you want to update an Invoice, you **must** make sure the Invoice's name is unique within the scope of the specified resource

        :calls: ``put /invoices/{id}``
        :param int id: Unique identifier of an Invoice.
        :param tuple *args: (optional) Single object representing Invoice resource which attributes should be updated.
        :param dict **kwargs: (optional) Invoice attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated Invoice resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Invoice are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, invoice = self.http_client.put("/invoices/{id}".format(id=id), body=attributes)
        return invoice



class EmployeeService(object):
    """
    :class:`fortnox.EmployeeService` is used by :class:`fortnox.Client` to make
    actions related to Employee resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for Employee to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = ['EmployeeId', 'FirstName', 'LastName']
    SERVICE = "Employee"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def list(self, **params):
        """
        Retrieve all Employees

        Returns all Employees available to the Company, according to the parameters provided

        :calls: ``get /employees``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of Employees.
        :rtype: list
        """

        _, _, employees = self.http_client.get("/employees", params=params)
        return employees

    def create(self, *args, **kwargs):
        """
        Create a Employee

        Creates a new Employee
        **Notice** the Employee's name **must** be unique within the scope of the resource_type

        :calls: ``post /customers``
        :param tuple *args: (optional) Single object representing Employee resource.
        :param dict **kwargs: (optional) Employee attributes.
        :return: Dictionary that support attriubte-style access and represents newely created Employee resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Employee are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items() if k in self.OPTS_KEYS_TO_PERSIST)
        attributes.update({'service': self.SERVICE})
        _, _, employee = self.http_client.post("/employees", body=attributes)
        return employee

    def retrieve(self, id):
        """
        Retrieve a single Employee

        Returns a single Employee according to the unique Employee ID provided
        If the specified Employee does not exist, this query returns an error

        :calls: ``get /employees/{id}``
        :param int id: Unique identifier of a employee.
        :return: Dictionary that support attriubte-style access and represent employee resource.
        :rtype: dict
        """
        _, _, employee = self.http_client.get("/employees/{id}".format(id=id))
        return employee

    def update(self, id, *args, **kwargs):
        """
        Update an Employee

        Updates an Employee's information
        If the specified Employee does not exist, this query will return an error
        **Notice** if you want to update an Employee, you **must** make sure the Employee's name is unique within the scope of the specified resource

        :calls: ``put /employees/{id}``
        :param int id: Unique identifier of a employee.
        :param tuple *args: (optional) Single object representing employee resource which attributes should be updated.
        :param dict **kwargs: (optional) employee attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated employee resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Employee are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, employee = self.http_client.put("/employees/{id}".format(id=id), body=attributes)
        return employee



class PredefinedVoucherSeriesService(object):
    """
    :class:`fortnox.PredefinedVoucherSeriesService` is used by :class:`fortnox.Client` to make
    actions related to Predefined Voucher Series resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for Predefined Voucher Series to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = []
    SERVICE = "PreDefinedVoucherSeries"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def list(self, **params):
        """
        Retrieve all PreDefined Vouchers

        Returns all Predefined Vouchers Series available to the Company, according to the parameters provided

        :calls: ``get /predefinedvoucherseries``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of PreDefined Voucher Series.
        :rtype: list
        """

        _, _, predefined_voucher_series_collection = self.http_client.get("/predefinedvoucherseries", params=params)
        return predefined_voucher_series_collection

    def retrieve(self, name):
        """
        Retrieve a single PreDefined Voucher Series

        Returns a single PreDefined Voucher Series according to the unique PreDefined Voucher Series ID provided
        If the specified PreDefined Voucher Series does not exist, this query returns an error

        :calls: ``get /predefinedvoucherseries/{name}``
        :param int id: Unique identifier of a PreDefined Voucher Series.
        :return: Dictionary that support attriubte-style access and represent PreDefined Voucher Series resource.
        :rtype: dict
        """
        _, _, predefined_voucher_series = self.http_client.get("/predefinedvoucherseries/{name}".format(name=name))
        return predefined_voucher_series

    def update(self, name, *args, **kwargs):
        """
        Update a PreDefined Voucher Series

        Updates a PreDefined Voucher's information
        If the specified PreDefined Voucher does not exist, this query will return an error
        **Notice** if you want to update a PreDefined Voucher, you **must** make sure the PreDefined Voucher's name is unique within the scope of the specified resource

        :calls: ``put /predefinedvoucherseries/{name}``
        :param int id: Unique identifier of a PreDefined Voucher.
        :param tuple *args: (optional) Single object representing PreDefined Voucher resource which attributes should be updated.
        :param dict **kwargs: (optional) PreDefined Voucher attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated PreDefined Voucher resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for PreDefined Voucher are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, predefined_voucher_series = self.http_client.put("/predefinedvoucherseries/{name}".format(name=name),
                                                               body=attributes)
        return predefined_voucher_series



class CompanyInformationService(object):
    """
    :class:`fortnox.CompanyInformationService` is used by :class:`fortnox.Client` to make
    actions related to Company Information resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for CompanyInformation to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = []
    SERVICE = "CompanyInformation"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def info(self, **params):
        """
        Retrieve Company Information

        Returns Company Information available to the Company, according to the parameters provided

        :calls: ``get /companyinformation``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of Company.
        :rtype: dict object
        """

        _, _, company = self.http_client.get("/companyinformation", params=params)
        return company



from decimal import Decimal


class Coercion(object):
    """
    Class used to make coertion between data returned from Base and data usable for developer
    """

    @staticmethod
    def to_decimal(value):
        """
        Coerce a value into a Decimal

        :param str value: a value to be coerced
        """
        return Decimal(str(value))

    @staticmethod
    def to_string(value):
        """
        Coerce a value into a String

        :param float value: a value to be coerced
        """
        return str(value)



class AssetService(object):
    """
    :class:`fortnox.AssetService` is used by :class:`fortnox.Client` to make
    actions related to Assets resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for Asset to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = ['Number', 'Description', 'TypeId', 'AcquisitionDate',
                            'AcquisitionStart', 'DepreciationFinal', 'AcquisitionValue']
    SERVICE = "Asset"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def list(self, **params):
        """
        Retrieve all Assets

        Returns all Assets available to the Company, according to the parameters provided

        :calls: ``get /assets``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of Assets.
        :rtype: list
        """

        _, _, assets = self.http_client.get("/assets", params=params)
        return assets

    def depreciation_list(self, to_date, **params):
        """
        Retrieve all deprecated assets

        Returns all deprecated assets available to the Company, according to the parameters provided

        :calls: ``get /assets/depreciations/{to_date}``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of deprecated assets.
        :rtype: list
        """

        _, _, deprecated_assets = self.http_client.get("/assets/depreciations/{to_date}".format(to_date=to_date),
                                                       params=params)
        return deprecated_assets[1:]

    def retrieve(self, id):
        """
        Retrieve a single Asset

        Returns a single Assets according to the unique Assets ID provided
        If the specified Asset does not exist, this query returns an error

        :calls: ``get /assets/{id}``
        :param int id: Unique identifier of an Asset.
        :return: Dictionary that support attriubte-style access and represent Asset resource.
        :rtype: dict
        """
        _, _, asset = self.http_client.get("/assets/{id}".format(id=id))
        return asset

    def create(self, *args, **kwargs):
        """
        Create an Asset

        Creates an new Asset
        **Notice** the Asset's name **must** be unique within the scope of the resource_type

        :calls: ``post /assets``
        :param tuple *args: (optional) Single object representing Asset resource.
        :param dict **kwargs: (optional) Asset attributes.
        :return: Dictionary that support attriubte-style access and represents newely created Asset resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Asset are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items() if k in self.OPTS_KEYS_TO_PERSIST)
        attributes.update({'service': self.SERVICE})
        _, _, asset = self.http_client.post("/assets", body=attributes)
        return asset

    def update(self, id, *args, **kwargs):
        """
        Update an Asset

        Updates an Asset's information
        If the specified Asset does not exist, this query will return an error
        **Notice** if you want to update an Asset, you **must** make sure the Asset's name is unique within the scope of the specified resource

        :calls: ``put /assets/{id}``
        :param int id: Unique identifier of an Asset.
        :param tuple *args: (optional) Single object representing Asset resource which attributes should be updated.
        :param dict **kwargs: (optional) Asset attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated Asset resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Asset are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, asset = self.http_client.put("/assets/{id}".format(id=id), body=attributes)
        return asset

    def destroy(self, id):
        """
        Delete an Asset

        Deletes an existing Asset
        If the specified Asset is assigned to any resource, we will remove this Asset from all such resources
        If the specified Asset does not exist, this query will return an error
        This operation cannot be undone

        :calls: ``delete /assets/{id}``
        :param int id: Unique identifier of an Asset.
        :return: True if the operation succeeded.
        :rtype: bool
        """

        status_code, _, _ = self.http_client.delete("/assets/{id}".format(id=id))
        return status_code == 200

    def depreciate(self, *args, **kwargs):
        """
        Depreciate Assets

        Updates an Asset's information
        If the specified Asset does not exist, this query will return an error
        **Notice** if you want to update an Asset, you **must** make sure the Asset's name is unique within the scope of the specified resource

        :calls: ``put /assets/depreciate``
        :param int id: Unique identifier of an Asset.
        :param tuple *args: (optional) Single object representing Asset resource which attributes should be updated.
        :param dict **kwargs: (optional) Asset attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated Asset resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Asset are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, asset = self.http_client.post("/assets/depreciate", body=attributes)
        return asset

    def write_up(self, id, *args, **kwargs):
        """
        Update an Asset

        Asset's Write Up information
        If the specified Asset does not exist, this query will return an error
        **Notice** if you want to update an Asset, you **must** make sure the Asset's name is unique within the scope of the specified resource

        :calls: ``put /assets/writeup/{id}``
        :param int id: Unique identifier of an Asset.
        :param tuple *args: (optional) Single object representing Asset resource which attributes should be updated.
        :param dict **kwargs: (optional) Asset attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated Asset resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Asset Write Up are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, asset = self.http_client.put("/assets/writeup/{id}".format(id=id), body=attributes)
        return asset

    def write_down(self, id, *args, **kwargs):
        """
        Write Down an Asset

        Updates an Asset's information
        If the specified Asset does not exist, this query will return an error
        **Notice** if you want to update an Asset, you **must** make sure the Asset's name is unique within the scope of the specified resource

        :calls: ``put /assets/writedown/{id}``
        :param int id: Unique identifier of an Asset.
        :param tuple *args: (optional) Single object representing Asset resource which attributes should be updated.
        :param dict **kwargs: (optional) Asset attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated Asset resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Asset Write Down are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, asset = self.http_client.put("/assets/writedown/{id}".format(id=id), body=attributes)
        return asset

    def scrap(self, id, *args, **kwargs):
        """
        Scrap Down an Asset

        Updates an Asset's information
        If the specified Asset does not exist, this query will return an error
        **Notice** if you want to update an Asset, you **must** make sure the Asset's name is unique within the scope of the specified resource

        :calls: ``put /assets/scrap/{id}``
        :param int id: Unique identifier of an Asset.
        :param tuple *args: (optional) Single object representing Asset resource which attributes should be updated.
        :param dict **kwargs: (optional) Asset attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated Asset resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Asset Scrap are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, asset = self.http_client.put("/assets/scrap/{id}".format(id=id), body=attributes)
        return asset

    def sell(self, id, *args, **kwargs):
        """
        Sell an Asset

        An Asset's sell information
        If the specified Asset does not exist, this query will return an error
        **Notice** if you want to update an Asset, you **must** make sure the Asset's name is unique within the scope of the specified resource

        :calls: ``put /assets/sell/{id}``
        :param int id: Unique identifier of an Asset.
        :param tuple *args: (optional) Single object representing Asset resource which attributes should be updated.
        :param dict **kwargs: (optional) Asset attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated Asset resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Asset are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, asset = self.http_client.put("/assets/sell/{id}".format(id=id), body=attributes)
        return asset



class AbsenceTransactionsService(object):
    """
    :class:`fortnox.AbsenceTransactions` is used by :class:`fortnox.Client` to make
    actions related to Absence Transactions resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for Absence Transactions to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = ['EmployeeId', 'CauseCode', 'Date', 'Extent']
    SERVICE = "AbsenceTransaction"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def list(self, **params):
        """
        Retrieve all AbsenceTransactions

        Returns all AbsenceTransactions available to the Company, according to the parameters provided

        :calls: ``get /absencetransactions``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of AbsenceTransactions.
        :rtype: list
        """

        _, _, absence_transactions = self.http_client.get("/absencetransactions", params=params)
        return absence_transactions

    def retrieve(self, employee_id, date, cause_code):
        """
        Retrieve a single AbsenceTransactions

        Returns a single AbsenceTransactions according to the unique AbsenceTransactions ID provided
        If the specified AbsenceTransactions does not exist, this query returns an error

        :calls: ``get /absencetransactions/{EmployeeId}/{Date}/{CauseCode}``
        :param int id: Unique identifier of a AbsenceTransactions.
        :return: Dictionary that support attriubte-style access and represent AbsenceTransactions resource.
        :rtype: dict
        """
        _, _, absence_transaction = self.http_client.get(
            "/absencetransactions/{EmployeeId}/{Date}/{CauseCode}".format(EmployeeId=employee_id, Date=date,
                                                                          CauseCode=cause_code))
        return absence_transaction

    def create(self, *args, **kwargs):
        """
        Create a AbsenceTransactions

        Creates a new AbsenceTransactions
        **Notice** the AbsenceTransaction's name **must** be unique within the scope of the resource_type

        :calls: ``post /absencetransactions``
        :param tuple *args: (optional) Single object representing AbsenceTransactions resource.
        :param dict **kwargs: (optional) AbsenceTransaction attributes.
        :return: Dictionary that support attriubte-style access and represents newely created Customer resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for AbsenceTransaction are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items() if k in self.OPTS_KEYS_TO_PERSIST)
        attributes.update({'service': self.SERVICE})
        _, _, absence_transaction = self.http_client.post("/absencetransactions", body=attributes)
        return absence_transaction

    def update(self, employee_id, date, cause_code, *args, **kwargs):
        """
        Update a AbsenceTransaction

        Updates a AbsenceTransaction's information
        If the specified Absence Transaction does not exist, this query will return an error
        **Notice** if you want to update a Absence Transaction, you **must** make sure the AbsenceTransaction's name is unique within the scope of the specified resource

        :calls: ``put /absencetransactions/{EmployeeId}/{Date}/{CauseCode}``
        :param int id: Unique identifier of a AbsenceTransaction.
        :param tuple *args: (optional) Single object representing Absence Transaction resource which attributes should be updated.
        :param dict **kwargs: (optional) AbsenceTransaction attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated AbsenceTransactions resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for AbsenceTransaction are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, absence_transaction = self.http_client.put(
            "/absencetransactions/{EmployeeId}/{Date}/{CauseCode}".format(EmployeeId=employee_id, Date=date,
                                                                          CauseCode=cause_code), body=attributes)
        return absence_transaction

    def destroy(self, employee_id, date, cause_code):
        """
        Delete a AbsenceTransaction

        Deletes an existing AbsenceTransaction
        If the specified Absence Transaction is assigned to any resource, we will remove this Absence Transaction from all such resources
        If the specified Absence Transaction does not exist, this query will return an error
        This operation cannot be undone

        :calls: ``delete /absencetransactions/{EmployeeId}/{Date}/{CauseCode}``
        :param int id: Unique identifier of a AbsenceTransactions.
        :return: True if the operation succeeded.
        :rtype: bool
        """

        status_code, _, _ = self.http_client.delete(
            "/absencetransactions/{EmployeeId}/{Date}/{CauseCode}".format(EmployeeId=employee_id, Date=date,
                                                                          CauseCode=cause_code))
        return status_code == 204



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyfortnox",
    version="0.0.5",
    author="Mahmudul Hasan",
    author_email="ikhtiarcse10ruet@gmail.com",
    description="A small package for Fortnox API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/xalien10/pyfortnox",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=["munch>=2.5.0", "requests>=2.21.0", "urllib3>=1.24.3"],
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



class ProjectService(object):
    """
    :class:`fortnox.ProjectService` is used by :class:`fortnox.Client` to make
    actions related to Project resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for Project to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = ['Description', ]
    SERVICE = "Project"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def list(self, **params):
        """
        Retrieve all Projects

        Returns all Projects available to the Company, according to the parameters provided

        :calls: ``get /projects``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of Projects.
        :rtype: list
        """

        _, _, projects = self.http_client.get("/projects", params=params)
        return projects

    def retrieve(self, number):
        """
        Retrieve a single project

        Returns a single project according to the unique project ID provided
        If the specified project does not exist, this query returns an error

        :calls: ``get /projects/{number}``
        :param int id: Unique identifier of a project.
        :return: Dictionary that support attriubte-style access and represent project resource.
        :rtype: dict
        """
        _, _, project = self.http_client.get("/projects/{number}".format(number=number))
        return project

    def create(self, *args, **kwargs):
        """
        Create a project

        Creates a new project
        **Notice** the project's name **must** be unique within the scope of the resource_type

        :calls: ``post /projects``
        :param tuple *args: (optional) Single object representing project resource.
        :param dict **kwargs: (optional) project attributes.
        :return: Dictionary that support attriubte-style access and represents newely created project resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Project are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items() if k in self.OPTS_KEYS_TO_PERSIST)
        attributes.update({'service': self.SERVICE})
        _, _, project = self.http_client.post("/projects", body=attributes)
        return project

    def update(self, number, *args, **kwargs):
        """
        Update a Project

        Updates a Project's information
        If the specified Project does not exist, this query will return an error
        **Notice** if you want to update a Project, you **must** make sure the Project's name is unique within the scope of the specified resource

        :calls: ``put /projects/{number}``
        :param int id: Unique identifier of a Project.
        :param tuple *args: (optional) Single object representing Project resource which attributes should be updated.
        :param dict **kwargs: (optional) Project attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated Project resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Project are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, customer = self.http_client.put("/projects/{number}".format(number=number), body=attributes)
        return customer

    def destroy(self, number):
        """
        Delete a project

        Deletes an existing project
        If the specified project is assigned to any resource, we will remove this project from all such resources
        If the specified project does not exist, this query will return an error
        This operation cannot be undone

        :calls: ``delete /projects/{number}``
        :param int id: Unique identifier of a project.
        :return: True if the operation succeeded.
        :rtype: bool
        """

        status_code, _, _ = self.http_client.delete("/projects/{number}".format(number=number))
        return status_code == 204



import re

from fortnox.errors import ConfigurationError


class Configuration(object):
    """Base CRM client configuration :class:`Configuration <Configuration>` object.

    Used by :class:`HttpClient <HttpClient>` to send requests to Fortnox's servers.
    """

    URL_REGEXP = r'\b(?:(?:https?|http):\/\/|www\.)[-a-z0-9+&@#\/%?=~_|!:,.;]*[-a-z0-9+&@#\/%=~_|]'

    def __init__(self, **options):
        """
        :param str access_token: Personal access token.
        :param str base_url: (optional) Base url for the api. Default: ``https://api.fortnox.se``.
        :param bool verbose: (optional) Verbose/debug mode. Default: ``False``.
        :param int timeout: (optional) Connection and response timeout. Default: **30** seconds.
        """

        self.access_token = options.get('access_token')
        self.client_secret = options.get('client_secret')
        self.base_url = options['base_url'] if 'base_url' in options else 'https://api.fortnox.se'
        self.timeout = options['timeout'] if 'timeout' in options else 30

    def validate(self):
        """Validates whether a configuration is valid.

        :rtype: bool
        :raises ConfigurationError: if no ``access_token`` provided.
        :raises ConfigurationError: if provided ``access_token`` is invalid - contains disallowed characters.
        :raises ConfigurationError: if provided ``access_token`` is invalid - has invalid length.
        :raises ConfigurationError: if provided ``base_url`` is invalid.
        """
        if self.access_token is None:
            raise ConfigurationError('No access token provided. '
                                     'Set your access token during client initialization using: '
                                     '"basecrm.Client(access_token= <YOUR_PERSONAL_ACCESS_TOKEN>)"')

        if re.search(r'\s', self.access_token):
            raise ConfigurationError('Provided access token is invalid '
                                     'as it contains disallowed characters. '
                                     'Please double-check you access token.')

        # if len(self.access_token) != 64:
        #     raise ConfigurationError('Provided access token is invalid '
        #                              'as it has invalid length. '
        #                              'Please double-check your access token.')

        if not self.base_url or not re.match(self.URL_REGEXP, self.base_url):
            raise ConfigurationError('Provided base url is invalid '
                                     'as it not a valid URI. '
                                     'Please make sure it incldues the schema part, '
                                     'both http and https are accepted, '
                                     'and the hierarchical part')

        return True



class AccountsService(object):
    """
    :class:`fortnox.AccountsService` is used by :class:`fortnox.Client` to make
    actions related to <specific-service> resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for Accounts to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = ['Number', 'Description']
    SERVICE = "Account"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def list(self, **params):
        """
        Retrieve all Accounts

        Returns all Accounts available to the Company, according to the parameters provided

        :calls: ``get /accounts``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of Customers.
        :rtype: list
        """

        _, _, accounts = self.http_client.get("/accounts", params=params)
        return accounts

    def retrieve(self, id):
        """
        Retrieve a single Accounts

        Returns a single Account according to the unique Account ID provided
        If the specified Account does not exist, this query returns an error

        :calls: ``get /accounts/{id}``
        :param int id: Unique identifier of a Account.
        :return: Dictionary that support attriubte-style access and represent Accounts resource.
        :rtype: dict
        """
        _, _, account = self.http_client.get("/accounts/{id}".format(id=id))
        return account

    def create(self, *args, **kwargs):
        """
        Create a Account

        Creates a new Accounts
        **Notice** the Account's name **must** be unique within the scope of the resource_type

        :calls: ``post /accounts``
        :param tuple *args: (optional) Single object representing Account resource.
        :param dict **kwargs: (optional) Account attributes.
        :return: Dictionary that support attriubte-style access and represents newely created Account resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Account are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items() if k in self.OPTS_KEYS_TO_PERSIST)
        attributes.update({'service': self.SERVICE})
        _, _, account = self.http_client.post("/accounts", body=attributes)
        return account

    def update(self, id, *args, **kwargs):
        """
        Update a Account

        Updates a Account's information
        If the specified Account does not exist, this query will return an error
        **Notice** if you want to update a Account, you **must** make sure the Account's name is unique within the scope of the specified resource

        :calls: ``put /accounts/{id}``
        :param int id: Unique identifier of a Account.
        :param tuple *args: (optional) Single object representing Accounts resource which attributes should be updated.
        :param dict **kwargs: (optional) account attributes to update.
        :return: Dictionary that support attriubte-style access and represents updated Accounts resource.
        :rtype: dict
        """

        if not args and not kwargs:
            raise Exception('attributes for Account are missing')

        attributes = args[0] if args else kwargs
        attributes = dict((k, v) for k, v in attributes.items())
        attributes.update({'service': self.SERVICE})
        _, _, account = self.http_client.put("/accounts/{id}".format(id=id), body=attributes)
        return account



class AccountChartsService(object):
    """
    :class:`fortnox.AccountChartsService` is used by :class:`fortnox.Client` to make
    actions related to Account Charts resource.

    Normally you won't instantiate this class directly.
    """

    """
    Allowed attributes for AccountCharts to send to Fortnox backend servers.
    """
    OPTS_KEYS_TO_PERSIST = []
    SERVICE = "AccountChart"

    def __init__(self, http_client):
        """
        :param :class:`fortnox.HttpClient` http_client: Pre configured high-level http client.
        """

        self.__http_client = http_client

    @property
    def http_client(self):
        return self.__http_client

    def list(self, **params):
        """
        Retrieve all Account Charts

        Returns all Account Chart available to the Company, according to the parameters provided

        :calls: ``get /accountcharts``
        :param dict params: (optional) Search options.
        :return: List of dictionaries that support attriubte-style access, which represent collection of Account Charts.
        :rtype: list
        """

        _, _, account_charts = self.http_client.get("/accountcharts", params=params)
        return account_charts
